"""
Module to define Float classes to calculate float numbers.
"""

from collections.abc import Iterable
import math
import osvg.float
import osvg.helper


class MathFloat(osvg.float.RefFloat):
    """
    Base class for all "calculated" Float classes
    """

    def __init__(
        self,
        float_ref: osvg.float.Float,
    ) -> None:
        super().__init__(float_ref=float_ref)
        self._value = None

    def fill_cache(self) -> None:
        """
        Calculate the value for the cache
        """
        raise NotImplementedError

    @property
    def value(self) -> float:
        if self._value is None:
            self.fill_cache()
        return self._value

    def my_input_changed(self):
        self._value = None
        return super().my_input_changed()


class InvertedSign(MathFloat):
    """
    Invert the sign of the referenced Float
    """

    def fill_cache(self) -> None:
        self._value = -self.float_ref.value


class Cos(MathFloat):
    """
    Cosinus value of an angle in degree
    """

    def fill_cache(self):
        self._value = math.cos(osvg.helper.radians(self.float_ref.value))


class Sin(MathFloat):
    """
    Sinus value of an angle in degree
    """

    def fill_cache(self):
        self._value = math.sin(osvg.helper.radians(self.float_ref.value))


class Cot(MathFloat):
    """
    Cotangent value of an angle in degree
    """

    def fill_cache(self):
        v = osvg.helper.radians(self.float_ref.value)
        self._value = math.cos(v) / math.sin(v)


class Tan(MathFloat):
    """
    Tangent value of an angle in degree
    """

    def fill_cache(self):
        v = osvg.helper.radians(self.float_ref.value)
        self._value = math.sin(v) / math.cos(v)


class CalculatedList(MathFloat):
    """
    Base class for all calculated Floats with a list of Float objects
    """

    def __init__(
        self,
        *args: Iterable[osvg.float.Float | float | int],
    ) -> None:
        super().__init__(float_ref=None)
        if not args:
            raise ValueError(f"{self.__class__} needs at least on argument")
        for v in args:
            if not issubclass(type(v), osvg.float.Float):
                v = osvg.float.Const(v)
            self.add_input(v)

    @property
    def floats(self) -> list[float]:
        """
        Get tuple of all float input object values
        """
        return tuple(x.value for x in self.inputs)

    def fill_cache(self) -> None:
        raise NotImplementedError


class CalculatedKwargs(MathFloat):
    """
    Base class for all calculated Floats with keyword Float objects arguments
    """

    def __init__(
        self,
        **kwargs: osvg.float.Float | float | int,
    ) -> None:
        super().__init__(float_ref=None)
        self.__dict__.update(kwargs)
        for k, v in kwargs.items():
            if not issubclass(type(v), osvg.float.Float):
                v = osvg.float.Const(v)
                kwargs[k] = v
            self.add_input(v)
        self.__dict__.update(kwargs)

    def fill_cache(self) -> None:
        raise NotImplementedError


class FunctFloat(CalculatedList):
    """
    Class for customized calculation of Floats by passing a external function
    """

    def __init__(
        self,
        args: Iterable[osvg.float.Float | float | int],
        funct: "function",
    ) -> None:
        super().__init__(*args)
        self.funct = funct

    def fill_cache(self) -> None:
        args = self.floats
        self._value = self.funct(*args)


class Prod(CalculatedList):
    """
    Multiply all given Floats.
    """

    def fill_cache(self) -> None:
        """
        Product of all given values
        """
        result = 1
        for v in self.floats:
            result = result * v
        self._value = result


class Sum(CalculatedList):
    """
    Sum up reference value and one or more other numbers or values.
    """

    def fill_cache(self) -> None:
        """
        Sum of all given values
        """
        self._value = sum(self.floats)


class Min(CalculatedList):
    """
    Get the lowest value of all given Floats.
    """

    def fill_cache(self) -> None:
        """
        Lowest values of all given values
        """
        self._value = min(self.floats)


class LowestAbs(CalculatedList):
    """
    Return the floating number with the lowest absolute value.
    """

    def fill_cache(self) -> None:
        """
        Lowest absolute value of all given values
        """
        floats = self.floats
        if not floats:
            raise RuntimeError("At least one input required")
        lowest_value = floats[0]
        for v in floats[1:]:
            if abs(v) < abs(lowest_value):
                lowest_value = v
        self._value = lowest_value


class Max(CalculatedList):
    """
    Get the highest value of all given Floats.
    """

    def fill_cache(self) -> None:
        """
        Highest values of all given values
        """
        self._value = max(self.floats)


class PercentOf(CalculatedKwargs):
    """
    Calculate percentage of a value
    """

    def __init__(
        self,
        reference: osvg.float.Float,
        percentage: int | float | osvg.float.Float,
        modulo: bool = False,
    ) -> None:
        super().__init__(reference=reference, percentage=percentage)
        self._modulo = False
        self.modulo = modulo

    @property
    def modulo(self) -> bool:
        """
        Get modulo value
        """
        return self._modulo

    @modulo.setter
    def modulo(self, _bool: bool) -> None:
        self._modulo = _bool
        self.input_changed()

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        if self.modulo:
            self._value = self.reference.value / 100 * (self.percentage.value % 100)
        else:
            self._value = self.reference.value / 100 * self.percentage.value



"""
Module to define the basic Float classes.
"""

from collections.abc import Iterable


class Float:
    """
    An Float is an object that has other Float objects as
    inputs and output.
    With Float you can build a loop-free directed graph for calculating
    dynamically float numbers.
    """

    str_round = 3

    def __init__(
        self, inputs: list["Float"] | None = None, outputs: list["Float"] | None = None
    ) -> None:
        self.inputs = []
        self.outputs = []
        if inputs is not None:
            for i in inputs:
                self.add_input(i)
        if outputs is not None:
            for o in outputs:
                self.add_output(o)

    @staticmethod
    def verify_other(other: any, class_check: bool = True):
        """
        Verify if other's class is a subclass to IOR
        """
        if class_check and not issubclass(type(other), Float):
            raise TypeError("Other's class is not a subclass to Float")

    @property
    def input_history(self) -> set["Float"]:
        """
        Get all inputs of all inputs recursively
        """
        input_history = set(self.inputs)
        for i in input_history:
            input_history.union(i.input_history)
        return input_history

    def add_input(self, other: "Float") -> None:
        """
        Add other Float object as a input
        """
        self.verify_other(other)
        if self in other.input_history:
            raise RuntimeError("Input loop detected")
        self.inputs.append(other)
        other.outputs.append(self)
        self.input_changed()

    def remove_input(self, other: "Float") -> None:
        """
        Remove other Float object from inputs
        """
        self.inputs.remove(other)
        other.outputs.remove(self)
        self.input_changed()

    def add_output(self, other: "Float") -> None:
        """
        Add other Float object as a output
        """
        self.verify_other(other)
        other.add_input(self)

    def remove_output(self, other: "Float") -> None:
        """
        Remove other Float object from outputs
        """
        other.remove_input(self)

    def replace_by(self, other: "Float") -> "Float":
        """
        - Remove all my inputs
        - Move all my outputs to other's outputs
        - Return other
        """
        if self == other:
            raise RuntimeError("Cannot replace myself")
        for i in self.inputs:
            self.remove_input(i)
        for o in self.outputs:
            other.add_output(o)
            self.remove_output(o)
        return other

    def my_input_changed(self):
        """
        Perform actions on input change
        """

    def input_changed(self):
        """
        Inform myself and all outputs about an input change
        """
        self.my_input_changed()
        for o in self.outputs:
            o.input_changed()

    @property
    def value(self) -> None:
        """
        Get the value of the reference Float object
        """
        raise NotImplementedError

    def __str__(self) -> str:
        v = round(self.value, self.str_round)
        if v == 0:
            v = 0
        return f"{v:g}"


class Const(Float):
    """
    Float with an attribute 'value' as a floating number
    """

    def __init__(self, value: float) -> None:
        super().__init__()
        self._value = float(value)

    @property
    def value(self) -> float:
        return self._value

    @value.setter
    def value(self, value: float | int) -> None:
        self._value = value
        self.input_changed()


class RefFloat(Float):
    """
    Is always an intermediate "input" to another Float object
    """

    def __init__(self, float_ref: Float = None):
        super().__init__()
        self._float_ref = None
        if float_ref is not None:
            self.float_ref = float_ref

    @property
    def float_ref(self) -> Float:
        """
        Get the reference Float object
        """
        return self._float_ref

    @float_ref.setter
    def float_ref(self, float_ref: Float):
        if float_ref is None:
            if self._float_ref is not None:
                self.remove_input(self._float_ref)
            self._float_ref = None
            return
        if not issubclass(type(float_ref), Float):
            raise TypeError("Float reference must be a Float object")
        if self._float_ref is not None:
            self._float_ref = self._float_ref.replace_by(float_ref)
        else:
            self._float_ref = float_ref
        self.add_input(self._float_ref)

    @property
    def value(self) -> float:
        if self.float_ref is None:
            return None
        return self.float_ref.value


class FloatProperty:
    """
    Property descriptor for Float objects.
    It automatically creates an intermediate Float object,
    which uses the given Float object as input.
    Or, if input is a python native float or int, it creates
    a Float object with that value and use it as the input.
    """

    def __init__(self):
        self.instances = {}

    def __get__(self, instance, owner):
        return self.instances[instance]

    def __set__(self, instance, f):
        if f is not None and not issubclass(type(f), Float):
            f = Const(f)
        if instance not in self.instances:
            self.instances[instance] = RefFloat(f)
        else:
            self.instances[instance].float_ref = f


class FloatList:
    """
    List of Float values
    """

    name = "values"

    def __init__(self, values: Iterable[int | float | Float] = None) -> None:
        self._values = []
        if values is not None:
            self.values = values

    @property
    def values(self) -> list[Float]:
        """
        Get value list
        """
        return self._values

    @values.setter
    def values(self, values: Iterable[int | float | Float]) -> None:
        """
        Verify and set values
        """
        if not isinstance(values, list):
            raise TypeError(f"{self.name} must be a list of float values")
        for value in values:
            if not issubclass(type(value), Float):
                value = Const(value)
            self._values.append(RefFloat(float_ref=value))

    def __str__(self) -> str:
        if not self.values:
            return ""
        return ",".join([str(value) for value in self.values])

    def copy(self):
        """
        Return an object with the same float values as values
        """
        return self.__class__(values=[v.float_ref for v in self.values])

    @property
    def isdefault(self) -> bool:
        """
        Check, if object's value is not set.
        """
        return self.values == []

    def equals(self, other) -> bool:
        """
        Check, if other's float value is equal to self's float value
        """
        for i, v in enumerate(self.values):
            if v.float_ref.value != other.values[i].float_ref.value:
                return False
        return True



"""
Module to define position objects.
"""

# pylint: disable=missing-function-docstring

import math
from collections.abc import Iterable
import osvg.float
import osvg.float_math
import osvg.helper


class Position:
    """
    Class to define a point within in the two-dimensional drawing.
    'x' is the floating point value on the x axis with
    x=0 as the most left point and x>0 is moving to the right.
    'y' is the floating point value on the y axis with
    y=0 as the most upper point and y>0 is moving down.

    A PositionObject is a linkable object.
    """

    # pylint: disable=too-few-public-methods

    str_round = 3
    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()

    def __init__(
        self,
        x: osvg.float.Float | float = 0.0,
        y: osvg.float.Float | float = 0.0,
    ) -> None:
        self.x = x
        self.y = y

    def __str__(self):
        """
        Get the SVG string
        """
        return f"{self.x},{self.y}"


class PositionProperty:
    """
    Property descriptor for Position objects
    """

    def __init__(self):
        self.instances = {}

    def __get__(self, instance, owner):
        return self.instances[instance]

    def __set__(
        self, instance, p: Position | Iterable[int | float | osvg.float.Float] = None
    ):
        if p is None:
            x = 0
            y = 0
        elif issubclass(type(p), Position):
            x = p.x
            y = p.y
        elif isinstance(p, Iterable):
            x = p[0]
            y = p[1]
        else:
            raise TypeError("Position object required")
        if instance not in self.instances:
            self.instances[instance] = Position()
        self.instances[instance].x = x
        self.instances[instance].y = y


class ShiftedPosition(Position):
    # pylint: disable=too-few-public-methods
    """
    A relative position object where the coordinates are shifted
    to the reference by values.
    """
    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    origin = PositionProperty()
    x_shift = osvg.float.FloatProperty()
    y_shift = osvg.float.FloatProperty()

    def __init__(
        self,
        origin: Position,
        x_shift: osvg.float.Float | float = 0,
        y_shift: osvg.float.Float | float = 0,
    ) -> None:
        super().__init__()
        self.origin = origin
        self.x_shift = x_shift
        self.y_shift = y_shift
        self.x = osvg.float_math.Sum(self.origin.x, self.x_shift)
        self.y = osvg.float_math.Sum(self.origin.y, self.y_shift)


class Distance(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    distance between two Positions
    """

    def __init__(self, a: Position, b: Position) -> None:
        super().__init__(ax=a.x, ay=a.y, bx=b.x, by=b.y)

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        self._value = math.dist(
            [self.ax.value, -self.ay.value],
            [self.bx.value, -self.by.value],
        )


class AngleDegree(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    angle from "a" to another position "b" in radians
    """

    def __init__(self, a: Position, b: Position) -> None:
        super().__init__(ax=a.x, ay=a.y, bx=b.x, by=b.y)

    def fill_cache(self) -> None:
        # pylint: disable=no-member
        dx = self.bx.value - self.ax.value
        dy = self.ay.value - self.by.value  # due to the inverted y-axis
        self._value = osvg.helper.degrees(math.atan2(dy, dx))


class XonCircle(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    x value of a position on a circle.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(self, angle: osvg.float.Float, radius: osvg.float.Float) -> None:
        super().__init__(angle=angle, radius=radius)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value = (
            math.cos(osvg.helper.radians(self.angle.value)) * self.radius.value
        )


class YonCircle(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated osvg.float.Float object representing the
    y value of a position on a circle.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, angle: osvg.float.Float, radius: osvg.float.Float) -> None:
        super().__init__(angle=angle, radius=radius)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value = (
            math.sin(osvg.helper.radians(self.angle.value)) * -1 * self.radius.value
        )


class XonEllipse(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    x value of a position on a ellipse.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(
        self,
        angle: osvg.float.Float,
        radius_x: osvg.float.Float,
        radius_y: osvg.float.Float,
    ) -> None:
        super().__init__(angle=angle, radius_x=radius_x, radius_y=radius_y)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        self._value = abs(
            self.radius_x.value
            * self.radius_y.value
            / math.sqrt(
                self.radius_y.value**2
                + self.radius_x.value**2
                * math.tan(osvg.helper.radians(self.angle.value)) ** 2
            )
        )
        if 90 < self.angle.value % 360 < 270:
            self._value = -1 * self._value


class YonEllipse(osvg.float_math.CalculatedKwargs):
    """
    Class to provide a calculated Float object representing the
    y value of a position on a ellipse.
    Note: angle=0 is twelve o'clock and it goes clockwise.
    """

    def __init__(
        self,
        angle: osvg.float.Float,
        radius_x: osvg.float.Float,
        radius_y: osvg.float.Float,
    ) -> None:
        super().__init__(angle=angle, radius_x=radius_x, radius_y=radius_y)

    def fill_cache(self) -> float:
        # pylint: disable=no-member
        tan = math.tan(osvg.helper.radians(self.angle.value))
        if tan == 0:
            self._value = 0
        else:
            self._value = abs(
                self.radius_x.value
                * self.radius_y.value
                / math.sqrt(
                    self.radius_x.value**2
                    + self.radius_y.value**2
                    / math.tan(osvg.helper.radians(self.angle.value)) ** 2
                )
            )
            if self.angle.value % 360 > 180:
                self._value = -1 * self._value


class RotatedPosition(Position):
    """
    Class to define a point rotated around another point by an angle.
    'reference' is the Position which should be shifted.
    'center' is the center of the circle with 'parent' on the cirle track.

    Note:
    - angle goes clockwise
    - angle is delta to current angle 'reference' -> 'parent'
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    origin = PositionProperty()
    center = PositionProperty()
    angle = osvg.float.FloatProperty()

    def __init__(
        self,
        origin: Position,
        center: Position,
        angle: osvg.float.Float | float = 0,
    ) -> None:
        super().__init__()
        self.origin = origin
        self.center = center
        self.angle = angle

        distance = Distance(self.center, self.origin)
        origin_angle = AngleDegree(self.center, self.origin)
        new_angle = osvg.float_math.Sum(self.angle, origin_angle)

        # Calc x
        origin_shift_x = XonCircle(angle=origin_angle, radius=distance)
        new_shift_x = XonCircle(angle=new_angle, radius=distance)
        self.x = osvg.float_math.Sum(
            self.origin.x, osvg.float_math.InvertedSign(origin_shift_x), new_shift_x
        )

        # Calc y
        origin_shift_y = YonCircle(angle=origin_angle, radius=distance)
        new_shift_y = YonCircle(angle=new_angle, radius=distance)
        self.y = osvg.float_math.Sum(
            self.origin.y, osvg.float_math.InvertedSign(origin_shift_y), new_shift_y
        )


class PolarShiftedPosition(Position):
    """
    Class to define a position which is described by a polar coordinate
    with the reference position as center.
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    origin = PositionProperty()
    angle = osvg.float.FloatProperty()
    distance = osvg.float.FloatProperty()
    x_shift = osvg.float.FloatProperty()
    y_shift = osvg.float.FloatProperty()

    def __init__(
        self,
        origin: Position,
        angle: osvg.float.Float | float = 0,
        distance: osvg.float.Float | float = 0,
    ) -> None:
        super().__init__()
        self.origin = origin
        self.angle = angle
        self.distance = distance

        x_shift = osvg.float_math.Prod(
            osvg.float_math.Cos(self.angle),
            self.distance,
        )
        self.x = osvg.float_math.Sum(origin.x, x_shift)
        y_shift = osvg.float_math.Prod(
            osvg.float_math.InvertedSign(osvg.float_math.Sin(self.angle)),
            self.distance,
        )
        self.y = osvg.float_math.Sum(origin.y, y_shift)


class RayCrossCirclePosition(Position):
    """
    Class to define a point at which a ray from
    circle center to a position crosses the circle.
    """

    # pylint: disable=too-few-public-methods

    x = osvg.float.FloatProperty()
    y = osvg.float.FloatProperty()
    center = PositionProperty()
    position = PositionProperty()
    radius = osvg.float.FloatProperty()
    x_shift = osvg.float.FloatProperty()
    y_shift = osvg.float.FloatProperty()

    def __init__(
        self,
        center: Position,
        position: Position,
        radius: osvg.float.Float | float,
    ) -> None:
        super().__init__()
        self.center = center
        self.position = position
        self.radius = radius

        angle = AngleDegree(a=self.center, b=self.position)
        self.x = osvg.float_math.Sum(
            center.x,
            XonCircle(angle=angle, radius=self.radius),
        )
        self.y = osvg.float_math.Sum(
            center.y,
            YonCircle(angle=angle, radius=self.radius),
        )
